import os
import numpy as np
import pylab as plt

"""
TODO: also goals must be classes
"""

class Parameter(object):
    """
    A Parameter
    """
    name = None
    min_value = None
    max_value = None
    values = None
    count = None
    index = None
    
    def __init__(self, name):
        self.name = name

    def __str__(self):
        retstr = 'Parameter: '
        retstr += '\t' + self.name
        return retstr
    
    def is_nominal(self):
        return self.count is not None

    def real_value(self, vals):
        if self.is_nominal():  # nominal parameter
            best_id = vals[self.index:self.index + self.count].argmax()
            return self.values[best_id]
        else:
            return vals[self.index] * (self.max_value - self.min_value) + self.min_value

class Element(object):
    """
    An element of NSGA-II
    """
    rank = None
    params = None
    results = None
    
    def __init__(self, arr, algo_info, num_goals, particle_size):
        self._set_results(arr[:num_goals])
        self._set_params(arr[num_goals:num_goals + particle_size], algo_info)
        self.rank = arr[-2]

    def __str__(self):
        retstr = 'Element (rank #%d)\n' % self.rank
        for p, v in self.params.iteritems():
            retstr += p + ' ' + str(v) + '\n'
        retstr += 'Results: '
        for res in self.results:
            retstr += str(res) + ' '
        retstr += '\n'
        return retstr
    
    def _set_params(self, vals, info):
        self.params = dict()
        for param in info.values():
            self.params[param.name] = param.real_value(vals)

    def _set_results(self, vals):
        self.results = vals

class Experiment(object):
    """
    Represents an experiment
    """
    
    def __init__(self, experiment_paths, parameter_info_file):
        # Read parameter information from C++ file
        self._parse_nsga_depso_file(parameter_info_file)
        # Read information about the experiment
        self._read_experiment_info(experiment_paths)
        # Read overall population
        self._load_final_population(experiment_paths)

    def _array_to_element(self, arr):
        """
        Convert a np.ndarray read from results to an element
        """
        return Element(arr, self._algorithm_info, self._num_goals, self._particle_size)

    def _dict_to_param(self, d):
        """
        Convert a dictionary to a parameter
        """
        p = Parameter(d['NAME'])
        if 'COUNT' in d:  # nominal parameter
            p.values = dict()
            p.count = int(d['COUNT'])
            for i in xrange(p.count):
                p.values[i] = eval(d[str(i)])
        else:  # numerical parameter
            p.min_value = float(d['MIN'])
            p.max_value = float(d['MAX'])
        p.index = int(d['INDEX'])
        return p

    def _get_run_info(self, path):
        """
        Returns information of a run, given path results
        """
        fp_info = open(os.path.join(path, 'nsga2_optimizer_parameters.out'), 'r')
        info = dict()
        suffix = ''
        for l in fp_info:
            ld = l.split()
            if ld[0] == 'PROBLEM':
                suffix = '_' + ld[1]
            else:
                info[ld[0] + suffix] = ld[1]
        fp_info.close()
        return info

    def _parse_nsga_depso_file(self, filename):
        """
        Recevies a configuration file and returns a dict containing the information
        regarding algorithms, parameters and parameter ranges.
        """
        algorithms = dict()
        particle_sizes = dict()
        with open(filename) as fp:
            all_lines = fp.read().splitlines()
        # Find valid algorithms
        for line in all_lines:
            if line.startswith('#define ALGO'):
                algo_name = eval(line.split()[-1])
                algorithms[algo_name] = dict()
        # For each algorithm, find parameters
        for algo_name, algo_data in algorithms.iteritems():
            for line in all_lines:
                if line.startswith('#define ' + algo_name + '_META_'):
                    _, param_data, value = line.split()
                    _, _, name, info = param_data.split('_')
                    if name == 'PARTICLE':
                        particle_sizes[algo_name] = int(value)
                        continue  # TODO: Do I have to keep memory of particle size?
                    if name not in algo_data:
                        algo_data[name] = dict()
                        algo_data[name]['NAME'] = name
                    algo_data[name][info] = value
        # Convert to Parameter object
        for algo_name, algo_data in algorithms.iteritems():
            for param_name, param in algo_data.iteritems():
                algo_data[param_name] = self._dict_to_param(param)

        self._algorithms_info = algorithms
        self._particle_sizes = particle_sizes

    def _load_final_population(self, paths):
        """
        Load final population of a (series of) experiments
        """
        raw_population = None
        for p in paths:
            if raw_population is None:
                raw_population = np.loadtxt(os.path.join(p, 'nsga2_final_pop.out'))
            else:
                raw_population = np.vstack((raw_population, np.loadtxt(os.path.join(p, 'nsga2_final_pop.out'))))
        population = list()
        for el in raw_population:
            population.append(self._array_to_element(el))
        self._raw_population = raw_population
        self._population = population

    def _read_experiment_info(self, paths):
        """
        Reads all experiment info and check they are all the same
        """
        info = None
        for p in paths:
            if info is None:
                info = self._get_run_info(p)
            else:
                tmp = self._get_run_info(p)
                if info != tmp:
                    print 'Error! All experiment must share the same information!'
                    exit(1)
        self._experiment_info = info
        self._num_goals = int(info['NUM_GOALS'])
        self._algorithm = info['ALGORITHM']
        self._multi_problem = int(info['MULTI_FUNCTION'])
        if self._algorithm not in self._algorithms_info:
            print 'The algorithm is %s but I only have information on %s' % (self._algorithm, ' '.join(self._algorithms_info.keys()))
            exit(1)
        self._algorithm_info = self._algorithms_info[self._algorithm]
        self._particle_size = self._particle_sizes[self._algorithm]

    def plot_2D(self, first_goal=0, second_goal=1, fitness_log=True, plot_now=True):
        info = self._experiment_info
        goal_str_0 = 'GOAL_%s' % first_goal
        goal_str_1 = 'GOAL_%s' % second_goal
        if fitness_log:
            if 'fitness' in info[goal_str_0]:
                plt.xscale('log')
            if 'fitness' in info[goal_str_1]:
                plt.yscale('log')
        plt.scatter(self._raw_population[:, first_goal], self._raw_population[:, second_goal])
        #plt.scatter(best_pop[:,0], best_pop[:,1], c='r', s=40)
        if self._multi_problem:
            plt.xlabel(info[goal_str_0] + ' ' + info['PROBLEM_NAME_%d' % first_goal])
            plt.ylabel(info[goal_str_1] + ' ' + info['PROBLEM_NAME_%d' % second_goal])
        else:
            plt.xlabel(info[goal_str_0] + ' ' + info['PROBLEM_NAME'])
            plt.ylabel(info[goal_str_1] + ' ' + info['PROBLEM_NAME'])
        if plot_now:
            plt.show()

    def plot_parameters(self, goal=0, normalized=True, fitness_log=True, plot_now=True):
        """
        Create some plots in which on one axis there is a parameter and
        on the other the goal
        """
        info = self._experiment_info
        goal_str = 'GOAL_%s' % goal
        for param_name, param in self._algorithm_info.iteritems():
            plt.figure(param_name)
            print param_name
            if fitness_log:
                if 'fitness' in info[goal_str]:
                    plt.yscale('log')
            if param.is_nominal():
                begin_id = self._num_goals + param.index
                x_data = self._raw_population[:, begin_id:begin_id + param.count].argmax(axis=1)
                plt.xticks(np.arange(param.count), [param.values[i] for i in xrange(param.count)])
            else:
                if normalized:
                    x_data = list()
                    for el in self._raw_population[:, self._num_goals:]:
                        x_data.append(param.real_value(el))
                    x_data = np.array(x_data)
                else:
                    x_data = self._raw_population[:, self._num_goals + param.index]
            plt.scatter(x_data, self._raw_population[:, goal])
            #plt.scatter(best_pop[:,0], best_pop[:,1], c='r', s=40)
            if self._multi_problem:
                plt.ylabel(info[goal_str] + ' ' + info['PROBLEM_NAME_%d' % goal])
            else:
                plt.ylabel(info[goal_str] + ' ' + info['PROBLEM_NAME'])
            plt.xlabel(param_name)
        if plot_now:
            plt.show()

    def print_best_elements(self, num=1):
        """
        Print best num elements according to each goal
        """
        if num < 1:
            print 'Error: num must be a positive number. Received:', num
            return
        for g in xrange(self._num_goals):
            print 'Best element for goal %d' % g  # when goals will be classes, print problem_name
            el_ids = self._raw_population[:, g].argsort()[:num]
            for el_id in el_ids:
                print self._population[el_id]
        
    def print_population(self, rank_max=None):
        """
        Print population
        """
        for el in self._population:
            if rank_max is None or el.rank <= rank_max:
                print el
