#include "de.h"
#include "random.h"
#include <limits>
#include <stdio.h>


void DECopyPosition(struct DEPSOPosition * to,const struct DEPSOPosition * from) {
    to->fitness = from->fitness;
    memcpy(to->x,from->x,sizeof(double) * to->size);
}

void DEInitPosition(struct DEPSOPosition * to,int size) {
    to->size = size;
    to->x = (double *) malloc(sizeof(double) * size);
}

void DEDestroyPosition(struct DEPSOPosition * to) {
    free(to->x);
}

void DEInitPositionArray(struct DEPSOPosition ** to,int size,int param_count) {
    int ii;
    (*to) = (struct DEPSOPosition *) malloc(sizeof(struct DEPSOPosition) * size);
    for (ii = 0; ii < size; ii++)
        DEInitPosition(&(*to)[ii],param_count);
}

void DEDestroyPositionArray(struct DEPSOPosition ** to,int size) {
    int ii;
    for (ii = 0; ii < size; ii++)
        DEDestroyPosition(&(*to)[ii]);
    free(*to);
}

void DEInitBoundaries(struct DEPSOBoundaries * to,int size,double initialMin,double initialMax) {
    int ii;
    to->min = (double *) malloc(size * sizeof(double));
    to->max = (double *) malloc(size * sizeof(double));
    to->size = size;

    for (ii = 0; ii < size; ii++) {
       to->min[ii] = initialMin;
       to->max[ii] = initialMax;
    }
}

void DEDestroyBoundaries(struct DEPSOBoundaries * to) {
    free(to->min);
    free(to->max);
}

void DESwapPositionPointers(struct DEPSOPosition ** one,struct DEPSOPosition ** other) {
     struct DEPSOPosition * temp = (*one);
     (*one) = (*other);
     (*other) = temp;
}

DEPSO_RESULTS DifferentialEvolution(
    const int population_size,
    int func_num,
    struct DECrossoverStrategy * strategy,
    int generation_count,
    const int evaluation_count,
    const StoppingCriterion stop_condition,  //0: goal fitness OR number of generation, 1: goal fitness OR use eval number, 2: use steady state OR number of generation, 3: use steady state OR eval number
    const float fitness_goal,
    const int problem_dimension,
    const struct DEPSOBoundaries * boundaries,
    struct DEPSOPosition * best_individual,
    const double min_improvement,    //the minimum improvement (%) for a state to not be considered steady state
    const int max_steady_state_generations  //
    ) {
    int i,j;
    double emin  = std::numeric_limits<double>::max(), prev_emin; // minimum fitness found so far
    int imin; // individual with fitness emin
    int gen, eval;

    struct DEPSOPosition * population;
    struct DEPSOPosition * new_population;
    struct DEPSOPosition * trial_population;
    struct DEPSOPosition best;
    struct DEPSOPosition prev_best;
    
	// struct containing final results
	DEPSO_RESULTS results;
	bool goal_reached = false;

    DEInitPositionArray(&population,population_size,problem_dimension);
    DEInitPositionArray(&new_population,population_size,problem_dimension);
    DEInitPositionArray(&trial_population,population_size,problem_dimension);
    DEInitPosition(&best,problem_dimension);
    DEInitPosition(&prev_best,problem_dimension);

    // Initialization
    // Right now this part is kept fairly simple and just generates
    // random numbers in the range [-initfac, +initfac]. You might
    // want to extend the init part such that you can initialize
    // each parameter separately.

    // spread initial population members
    for (i = 0; i < population_size; i++) {
        for (j = 0; j < problem_dimension; j++) {
            population[i].x[j] = alea_double(boundaries->min[j],boundaries->max[j]);
        }
    }
    test_func(population, problem_dimension, population_size, func_num);

    emin = population[0].fitness;
    imin = 0;
    for (i = 1; i < population_size; i++) {
        if (population[i].fitness < emin) {
            emin = population[i].fitness;
            imin = i;
        }
    }

    DECopyPosition(&best,&population[imin]);
    DECopyPosition(&prev_best,&best);

	// If I am considering evaluations
	if(stop_condition & STOP_EVALS)
		generation_count = evaluation_count / population_size;

    

    // Iteration loop
    gen = 1; // generation counter reset
    eval = population_size;
    int steady_state_count = 0;
    bool exit_asap = false;
    while (gen < generation_count && !exit_asap) {
        gen++;
        imin = 0;
        prev_emin = emin;

        // Crossover and mutation
        for (i = 0; i < population_size; i++) {
            DECopyPosition(&(trial_population[i]),&(population[i]));
            // crossover
            strategy->func(population_size,problem_dimension,&prev_best,
                population,&(trial_population[i]),i,strategy->data);

            // check boundaries
            for (j = 0; j < problem_dimension; j++) {
                if (trial_population[i].x[j] > boundaries->max[j])
                    trial_population[i].x[j] = boundaries->max[j];
                if (trial_population[i].x[j] < boundaries->min[j])
                    trial_population[i].x[j] = boundaries->min[j];
            }
        }
        // Evaluate trial population. Test how good these choice really were.
        test_func(trial_population, problem_dimension, population_size, func_num);
        eval += population_size;
        // Selection
        for (i = 0; i < population_size; i++) {
            // improved objective function value?
            if (trial_population[i].fitness <= population[i].fitness) {
                DECopyPosition(&new_population[i],&(trial_population[i]));
                // Was this a new minimum?
                if (trial_population[i].fitness < emin) {
                    // reset emin to new low...
                    emin = trial_population[i].fitness;
                    imin = i;
                    DECopyPosition(&best,&(trial_population[i]));
                }
            } else {
                // replace target with old value
                DECopyPosition(&new_population[i],&population[i]);
            }
        }

        DECopyPosition(&prev_best,&best);

        // swap population arrays. New generation becomes old one
        // TODO: you do not need new_population, can overwrite population
        DESwapPositionPointers(&new_population,&population);
        
        //check if goal has been reached
        if(stop_condition & STOP_SS) {
            if((prev_emin - emin)/prev_emin*100.0 < min_improvement) {
                if(++steady_state_count > max_steady_state_generations) {
                    goal_reached = true;
                    exit_asap = true;
                }
            } else {
                steady_state_count = 0;
            }
        } else if(emin < fitness_goal && !goal_reached) {
			goal_reached = true;
			results.convergence_gen = gen;
			results.convergence_evals = eval;
            if(stop_condition & STOP_FIT)
            {
                exit_asap = true;
            }
		}
        // try to save time
        if(emin == 0.0)
            exit_asap = true;
    }

    if (best_individual)
        DECopyPosition(best_individual,&best);

    DEDestroyPositionArray(&population,population_size);
    DEDestroyPositionArray(&new_population,population_size);
    DEDestroyPositionArray(&trial_population,population_size);
    DEDestroyPosition(&best);
    DEDestroyPosition(&prev_best);
    
    //save the results
    results.best_fitness = emin;
    if(not goal_reached)
    {
		results.convergence_gen = 1e9;
		results.convergence_evals = 1e9;//evaluation_count;
	}
    
    return results;
}
