#ifndef DE_H
#define DE_H

/// @file
/// Header for Differential Evolution evolutionary algorithm.

#include <string.h>
#include <stdlib.h>

#include "common.h"
#include "test_func.h"

/// Prototype for a crossover strategy.
/// @param[in] population_size the size of the population
/// @param[in] problem_dimension the number of variables of the problem.
/// @param[in] best the best individual in the population.
/// @param[in] population the population of population_size individual.
/// @param[in,out] individual the current individual, to be edited.
/// @param[in] individual_id the position of individual in population.
/// @param[in] data the custom data, as set in DECrossoverStrategy.data
// int population_size,int problem_dimension,
// struct DEPSOPosition * best,struct DEPSOPosition population[],
// struct DEPSOPosition * individual, <-- initialized with the current individual to be edited
// int individual_id, <-- id of individual in the population
// void * data
#define DE_CROSSOVER_STRATEGY(x) void (*x)(int,int,struct DEPSOPosition *,struct DEPSOPosition *,\
    struct DEPSOPosition * individual,int individual_id,void *)


/// A crossover strategy.
struct DECrossoverStrategy {
    /// The strategy function.
    DE_CROSSOVER_STRATEGY(func);
    /// The custom data.
    void * data;
};

/// Executes a Differential Evolution evolutionary algorithm.
/// @param[in] population_size the size of the population.
/// @param[in] fitness_function the fitness function.
/// @param[in] strategy the crossover strategy.
/// @param[in] generation_count the number of generations.
/// @param[in] evaluation_count the number of evaluations.
/// @param[in] use_evals_to_stop tells which is the stopping criterion.
/// @param[in] fitness_goal is the fitness value for which the problem is solved.
/// @param[in] problem_dimension the number of variables.
/// @param[in] boundaries the boundaries.
/// @param[in] best_individual if not NULL, the best solution found
///            will be copied here at the end of the execution.
///            If not NULL, MUST be already initialized.
/// @returns the fitness of the best individual.
DEPSO_RESULTS DifferentialEvolution(
    const int population_size,
    int func_num,
    struct DECrossoverStrategy * strategy,
    int generation_count,
    const int evaluation_count,
    const StoppingCriterion use_evals_to_stop,
    const float fitness_goal,
    const int problem_dimension,
    const struct DEPSOBoundaries * boundaries,
    struct DEPSOPosition * best_individual,
    const double min_improvement,    //the minimum improvement (%) for a state to not be considered steady state
    const int max_steady_state_generations  //
    );

/// Destroys an array of size DEPSOPosition s.
void DEDestroyPositionArray(struct DEPSOPosition ** to,int size);
/// Initializes an array of size DEPSOPosition s, with param_count variables.
void DEInitPositionArray(struct DEPSOPosition ** to,int size,int param_count);
/// Destroys a DEPSOPosition.
void DEDestroyPosition(struct DEPSOPosition * to);
/// Initializes a DEPSOPosition with size variables.
void DEInitPosition(struct DEPSOPosition * to,int size);
/// Copies a DEPSOPosition. The destination MUST be already initialized.
void DECopyPosition(struct DEPSOPosition * to,const struct DEPSOPosition * from);
/// Initializes a DEPSOBoundaries object.
/// @param[out] to the object.
/// @param[in] size the number of variables.
/// @param[in] initialMin the default value for field min. For exceptions,
///            write to the field directly.
/// @param[in] initialMax the default value for field max. For exceptions,
///            write to the field directly.
void DEInitBoundaries(struct DEPSOBoundaries * to,int size,double initialMin,double initialMax);
/// Destroys a DEPSOBoundaries object.
void DEDestroyBoundaries(struct DEPSOBoundaries * to);
/// Swaps two pointers to a position array.
void DESwapPositionPointers(struct DEPSOPosition ** one,struct DEPSOPosition ** other);

#endif // DE_H
