#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#include "random.h"
#include "pso.h"


void PSOCopyPosition(struct DEPSOPosition * to,const struct DEPSOPosition * from) {
    to->fitness = from->fitness;
    memcpy(to->x,from->x,sizeof(double) * to->size);
}

void PSOCopyVelocity(struct PSOVelocity * to,const struct PSOVelocity * from) {
    memcpy(to->v,from->v,sizeof(double) * to->size);
}

void PSOInitPosition(struct DEPSOPosition * to,int size) {
    to->size = size;
    to->x = (double *) malloc(sizeof(double) * size);
}

void PSOInitVelocity(struct PSOVelocity * to,int size) {
    to->size = size;
    to->v = (double *) malloc(sizeof(double) * size);
}

void PSODestroyPosition(struct DEPSOPosition * to) {
    free(to->x);
}

void PSODestroyVelocity(struct PSOVelocity * to) {
    free(to->v);
}

void PSOInitPositionArray(struct DEPSOPosition ** to,int size,int param_count) {
    int ii;
    (*to) = (struct DEPSOPosition *) malloc(sizeof(struct DEPSOPosition) * size);
    for (ii = 0; ii < size; ii++)
        PSOInitPosition(&(*to)[ii],param_count);
}

void PSOInitVelocityArray(struct PSOVelocity ** to,int size,int param_count) {
    int ii;
    (*to) = (struct PSOVelocity *) malloc(sizeof(struct PSOVelocity) * size);
    for (ii = 0; ii < size; ii++)
        PSOInitVelocity(&(*to)[ii],param_count);
}

void PSODestroyPositionArray(struct DEPSOPosition ** to,int size) {
    int ii;
    for (ii = 0; ii < size; ii++)
        PSODestroyPosition(&(*to)[ii]);
    free(*to);
}

void PSODestroyVelocityArray(struct PSOVelocity ** to,int size) {
    int ii;
    for (ii = 0; ii < size; ii++)
        PSODestroyVelocity(&(*to)[ii]);
    free(*to);
}

void PSOInitBoundaries(struct DEPSOBoundaries * to,int size,double initialMin,double initialMax) {
    int ii;
    to->min = (double *) malloc(size * sizeof(double));
    to->max = (double *) malloc(size * sizeof(double));
    to->size = size;

    for (ii = 0; ii < size; ii++) {
       to->min[ii] = initialMin;
       to->max[ii] = initialMax;
    }
}

void PSODestroyBoundaries(struct DEPSOBoundaries * to) {
    free(to->min);
    free(to->max);
}

DEPSO_RESULTS ParticleSwarmOptimization(
    const int population_size,
    const double w,const double cself,const double cglobal, // confidence coefficients
    int func_num,
    struct PSOBestFunction * best_function,
    int generation_count,
    const int evaluation_count,
    const StoppingCriterion stop_condition,
    const float fitness_goal,
    const int problem_dimension,
    const struct DEPSOBoundaries * boundaries,
    struct DEPSOPosition * best_individual,
    const double min_improvement,    //the minimum improvement (%) for a state to not be considered steady state
    const int max_steady_state_generations  //
    ) {
    int best;                    // Best of the best position (rank in the swarm)
    struct DEPSOPosition * best_X;   // Best positions found by each particle
    struct PSOVelocity * V;        // Velocities
    struct DEPSOPosition * X;        // Positions
    int iteration_counter;
    int ii,hh;
    int ibest;
    double result = 1e20, prev_result; // best fitness
    
	// struct containing final results
	DEPSO_RESULTS results;
	bool goal_reached = false;

    PSOInitPositionArray(&best_X,population_size,problem_dimension);
    PSOInitPositionArray(&X,population_size,problem_dimension);
    PSOInitVelocityArray(&V,population_size,problem_dimension);

    iteration_counter = 0;

    for (ii = 0; ii < population_size; ii++){  // Positions and velocities
        for (hh = 0;hh < problem_dimension;hh++){
            X[ii].x[hh] =  alea_double(boundaries->min[hh],boundaries->max[hh]);
            V[ii].v[hh] = (alea_double(boundaries->min[hh],boundaries->max[hh]) - X[ii].x[hh]) / 2.0; // Non uniform
            // V[s].v[d] = ( xmin[d]-xmax[d] )*(0.5-alea(0,1)); //Uniform. 2006-02-24
        }
    }
    
	// If I am considering evaluations
	if(stop_condition & STOP_EVALS)
		generation_count = evaluation_count / population_size;

    // First evaluations
    test_func(X, problem_dimension, population_size, func_num);
    for (ii = 0; ii < population_size; ii++ ){
        if(X[ii].fitness < result)
			result = X[ii].fitness;
        PSOCopyPosition(&best_X[ii],&X[ii]); // Best position = current one
    }

    int steady_state_count = 0;
    int eval = 0;
    bool exit_asap = false;
    do {
        if (best_function->global)
            ibest = best_function->global(population_size,best_X,best_function->data);

        //save the previous value
        prev_result = result;
        
        // The swarm MOVES
        for (ii = 0; ii < population_size; ii++ ){ // For each particle ...
            // .. find the local best
            if (best_function->local)
              ibest = best_function->local(population_size,ii,best_X,best_function->data);

            // ... compute the new velocity, and move
            for (hh = 0; hh < problem_dimension; hh++){
                V[ii].v[hh] = w * V[ii].v[hh] + alea_double(0,cself) * (best_X[ii].x[hh] - X[ii].x[hh]);
                V[ii].v[hh] += alea_double(0,cglobal) * (best_X[ibest].x[hh] - X[ii].x[hh]);
                X[ii].x[hh] = X[ii].x[hh] + V[ii].v[hh];
            }
            
            // ... interval confinement (keep in the box)
            for (hh = 0; hh < problem_dimension; hh++){
                if (X[ii].x[hh] < boundaries->min[hh]){
                    X[ii].x[hh] = boundaries->min[hh];
                }
                if (X[ii].x[hh] > boundaries->max[hh]){
                    X[ii].x[hh] = boundaries->max[hh];
                }
            }
        }
        // Evaluate swarm
        test_func(X, problem_dimension, population_size, func_num);
        
        // Update Bests
        for (ii = 0; ii < population_size; ii++ ){ // For each particle ...
			if(X[ii].fitness < result)
				result = X[ii].fitness;

            // ... update the best previous position
            if (X[ii].fitness < best_X[ii].fitness){
                PSOCopyPosition(&best_X[ii],&X[ii]);
            }
        }
        
        eval += population_size;
        
        if(stop_condition & STOP_SS) {
            if((prev_result - result)/prev_result*100.0 < min_improvement) {
                if(++steady_state_count > max_steady_state_generations) {
                    goal_reached = true;
                    exit_asap = true;
                }
            } else {
                steady_state_count = 0;
            }
        } else if(result < fitness_goal && !goal_reached) {
            goal_reached = true;
 			results.convergence_gen = iteration_counter;
 			results.convergence_evals = iteration_counter * population_size;
            if(stop_condition & STOP_FIT)
                exit_asap = true;
        }
        // try to save time
        if(result == 0.0)
            exit_asap = true;
		iteration_counter++;
    }
    while (iteration_counter < generation_count && !exit_asap);

    best = 0;
    for (ii = 1;ii < population_size; ii++) {
        if (best_X[ii].fitness < best_X[best].fitness)
            best = ii;
    }

    if (best_individual) {
        PSOCopyPosition(best_individual,&best_X[best]);
    }
    result = best_X[best].fitness;
    results.best_fitness = result;
    if(not goal_reached)
    {
        results.convergence_gen = 1e9;//iteration_counter;
        results.convergence_evals = 1e9;//eval;
 	}

    PSODestroyPositionArray(&best_X,population_size);
    PSODestroyPositionArray(&X,population_size);
    PSODestroyVelocityArray(&V,population_size);
  
    return results;
}
