#ifndef PSO_H
#define PSO_H

#include "common.h"
#include "test_func.h"

/// @file
/// Particle Swarm Optimization header.
/// A global best function. This function computes the best individual
/// in the swarm.
/// @param[in] population_size the size of the population.
/// @param[in] population the population, as array of population_size individuals.
/// @param[in,out] data the custom data, untouched,
///        as set in PSOBestFunction
/// @returns the index of the best individual
// int population_size, struct DEPSOPosition * population, void * data
#define PSO_GLOBAL_BEST_FUNCTION(x) int (* x)(int,struct DEPSOPosition *,void *)
/// A local best function. This function computes the best individual
/// in the swarm, given the current individual. If the information about current individual
/// is not needed, a global best function should be used for performance.
/// @param[in] population_size the size of the population.
/// @param[in] current_individual the current individual index
/// @param[in] population the population, as array of population_size individuals.
/// @param[in,out] data the custom data, untouched,
///        as set in PSOBestFunction
/// @returns the index of the best individual
// int population_size, int current_individual, struct DEPSOPosition * population, void * data
#define PSO_LOCAL_BEST_FUNCTION(x) int (* x)(int,int,struct DEPSOPosition *,void *)


/// Represents a velocity.
/// Internal use only, currently.
struct PSOVelocity {
    int size;
    double * v;
};


/// A Best Function, a function used to obtain the best individual
/// in the neighborhood of the current individual.
/// At least one between "global" and "local" MUST be not NULL.
struct PSOBestFunction {
    /// The global function will be called once for each global cycle and
    /// it returns the best individual of the swarm or
    /// it can store some data to "data" for further computations by
    /// "local" if it's also not NULL
    PSO_GLOBAL_BEST_FUNCTION(global);

    /// This function will be called once for each individual and for each cycle
    /// it returns the best individual in the neighbordhood
    PSO_LOCAL_BEST_FUNCTION(local);

    /// The custom data.
    void * data;
};

/// Executes the Particle Swarm Optimization evolutionary algorithm
/// to optimize a fitness function. This solve MINIMUM problems.
/// @param[in] population_size the number of particles.
/// @param[in] w the inertia of the particles.
/// @param[in] cself the attraction coefficient toward the best position
///        visited by the particle so far.
/// @param[in] cglobal the attraction coefficient toward the best position
///        visited by all the particles so far.
/// @param[in] fitness_function the fitness function (see PSOFitnessFunction).
/// @param[in] generation_count the number of generations.
/// @param[in] param_count the number of dimensions of the problem, i. e.
///        the number of real variables to be optimized.
/// @param[in] boundaries the boundaries of the function (see PSOBoundaries)
/// @param[out] best_individual if not NULL, will be filled with the best individual
///             found during optimization. If not NULL, it MUST be already
///             initialized, with the right size.
/// @returns the fitness of the best individual found.
DEPSO_RESULTS ParticleSwarmOptimization(
    const int population_size,
    const double w,const double cself,const double cglobal, // confidence coefficients
    int func_num,
    struct PSOBestFunction * best_function,
    int generation_count,
    const int evaluation_count,
    const StoppingCriterion use_evals_to_stop,
    const float fitness_goal,
    const int problem_dimension,
    const struct DEPSOBoundaries * boundaries,
    struct DEPSOPosition * best_individual,
    const double min_improvement,    //the minimum improvement (%) for a state to not be considered steady state
    const int max_steady_state_generations  //
    );

/// Internal use only.
void PSODestroyVelocityArray(struct PSOVelocity ** to,int size);
/// Destroys an array of size positions.
void PSODestroyPositionArray(struct DEPSOPosition ** to,int size);
/// Internal use only.
void PSOInitVelocityArray(struct PSOVelocity ** to,int size,int param_count);
/// Initializes an array of size positions, with param_count variables each.
void PSOInitPositionArray(struct DEPSOPosition ** to,int size,int param_count);
/// Internal use only.
void PSODestroyVelocity(struct PSOVelocity * to);
/// Destroys a position.
void PSODestroyPosition(struct DEPSOPosition * to);
/// Internal use only.
void PSOInitVelocity(struct PSOVelocity * to,int size);
/// Initializes a position with size variables.
void PSOInitPosition(struct DEPSOPosition * to,int size);
/// Internal use only.
void PSOCopyVelocity(struct PSOVelocity * to,const struct PSOVelocity * from);
/// Copies a position.
void PSOCopyPosition(struct DEPSOPosition * to,const struct DEPSOPosition * from);
/// Initializes a PSOBoundaries object.
/// @param[out] to the object.
/// @param[in] size the number of variables.
/// @param[in] initialMin the default value for field min. For exceptions,
///            write to the field directly.
/// @param[in] initialMax the default value for field max. For exceptions,
///            write to the field directly.
void PSOInitBoundaries(struct DEPSOBoundaries * to,int size,double initialMin,double initialMax);
/// Destroys a PSOBoundaries object.
void PSODestroyBoundaries(struct DEPSOBoundaries * to);

#endif // PSO_H
