#ifndef NSGA2_H_
#define NSGA2_H_

#include <nsga2/global.h>
#include <nsga2/random.h>

#include <vector>
#include <utility>
#include <iostream>
#include <fstream>
#include <string>

extern nsga2::random_gen rgen; // global common random generator

namespace nsga2 {

/// \brief A NSGA-II Optimization Process

/// It should work both with real-valued and binary individuals
class NSGA2 {
public:
    NSGA2();
    virtual ~NSGA2();

    void initialize() throw (nsga2exception);
    void advance();
    void evolve();

    /// Set seed of the random generator
    void set_seed(int s) {
        rgen.set_seed(s);
    };
    void set_function(individual_config::funcType f) {
        this->function = f;
    };
    void set_popfunction(individual_config::popFuncType f) {
        this->popFunction = f;
    };
    /// \brief Set crowding distance type
    
    /// \param crowd if it is true, the crowding distance will be computed over objectives,
    /// otherwise on the parameters
    void set_crowdobj(bool crowd) {
        this->crowd_obj = crowd;
    };
    /// Set name of experiment (used for output)
    /// \param n name of the folder that will contains the results
    /// at the end of execution
    void set_experiment_name(std::string n) {
		this->result_dir = n;
	};
    /// Set number of real variables (problem dimension)
    void set_nreal(int nreal) {
        this->nreal = nreal;
    };
    void set_nbin(int nbin) {
        this->nbin = nbin;
    };
    /// Set number of objectives
    void set_nobj(int nobj) {
        this->nobj = nobj;
    };
    /// Set significance value used in non-parameteric statistical test (if used).
    /// See set_comparison()
    void set_pvalue(float pvalue) {
        this->pvalue = pvalue;
    };
    /// Set number of repetitions used in non-parameteric statistical test (if used).
    /// See set_comparison()
    void set_nrep(int nrep) {
        this->nrep = nrep;
    };
    /// Set number of constraints
    void set_ncon(int ncon) {
        this->ncon = ncon;
    };
    /// Set Population Size
    void set_popsize(int popsize) {
        this->popsize = popsize;
    };
    /// Set number of generations
    void set_ngen(int ngen) {
        this->ngen = ngen;
    };
    /// Set number of generations between a report and the next
    void set_nreport(int nrep) {
	this->nreport = nrep;
    };
    /// Set probability of real crossover
    void set_pcross_real(double pcross_real) {
        this->pcross_real = pcross_real;
    };
    /// Set probability of binary crossover
    void set_pcross_bin(double pcross_bin) {
        this->pcross_bin = pcross_bin;
    };
    /// Set mutation probability for real variables
    void set_pmut_real(double pmut_real) {
        this->pmut_real = pmut_real;
    };
    /// Set mutation probability for binary variables
    void set_pmut_bin(double pmut_bin) {
        this->pmut_bin = pmut_bin;
    };
    /// Set distribution index for crossover
    void set_eta_c(double eta_c) {
        this->eta_c = eta_c;
    };
    /// Set distribution index for mutation
    void set_eta_m(double eta_m) {
        this->eta_m = eta_m;
    };
    void set_epsilon_c(double epsi_c) {
        this->epsilon_c = epsi_c;
    };
    /// Set number of bits for each binary variable
    void set_nbits(const std::vector<int>& nbits) {
        this->nbits = nbits;
    };
    /// Set limits of the search space for the real variables
    /// \param limits_realvar an array of couple in which each element represent the
    /// (minimum, maximum) value of the search space
    void set_limits_realvar(const std::vector< std::pair<double,double> >& limits_realvar) {
        this->limits_realvar = limits_realvar;
    };
    /// Set limits of binary variables
    void set_limits_binvar(const std::vector< std::pair<double,double> >& limits_binvar) {
        this->limits_binvar = limits_binvar;
    };
    /// Set the filename on which the backup of the current status will be saved (see save_backup())
    void set_backup_filename(const std::string& filename) {
	this->backupFilename = filename;
    }
    /// Set a different function to log results
    void set_custom_report_function(individual_config::popFuncType f) {
	this->reportFunction = f;
    }
    /// \brief Set comparison method
    
    /// Only 0 is supported. 0 means that fitnesses are treated just as normal fitnesses.
    /// If you put 1 or 2, an individual dominates another only if there is statistical evidence
    /// across all runs. It was an unsuccessful test that could be recovered someday.
    void set_comparison(int comparison_method) {
	this->comparison_method = comparison_method;
    }

    /// Save to file the current state
    void save_backup() const;
    /// \brief Load a previously unfinished run
    
    /// This class can recover an optimization problem that has been previously
    /// saved using save_backup()
    bool load_backup();

    /// Returns the path in which the results will be written
    std::string get_result_dir() const { return result_dir; }

private:
    // Parameters to be defined by the user
    int nreal;
    int nbin;
    int nobj;
    int nrep; // number of runs
    int ncon;
    int popsize;
    int ngen;
    int nreport;
    double pvalue; // statistical significance
    double pcross_real;
    double pcross_bin;
    double pmut_real;
    double pmut_bin;
    double eta_c;
    double eta_m;
    double epsilon_c;
    std::vector<int> nbits;
    std::vector< std::pair<double,double> > limits_realvar;
    // std::vector<double> min_realvar;
    // std::vector<double> max_realvar;
    // double *min_binvar;
    // double *max_binvar;
    std::vector< std::pair<double,double> > limits_binvar;
    // int choice; // to be added later, maybe.
    // int obj1;
    // int obj2;
    // int obj3;
    // int angle1;
    // int angle2;
    individual_config::funcType function;
    individual_config::popFuncType popFunction;
    individual_config::popFuncType reportFunction;
    int t;

    std::string backupFilename;

    bool crowd_obj;
    
    // Used to log results...
    std::ofstream fpt1;
    std::ofstream fpt2;
    std::ofstream fpt3;
    std::ofstream fpt4;
    std::ofstream fpt5;

    int nbinmut;
    int nrealmut;
    int nbincross;
    int nrealcross;
    int bitlength;
    /// Parent population
    population* parent_pop;
    /// Child population
    population* child_pop;
    /// Mixed population
    population* mixed_pop;
    /// How to say that an individual dominates the other: in some cases
    /// stronger assumptions than > are needed, such as statistical significance
    int comparison_method;

	std::string result_dir;

    void selection(population& oldpop, population& newpop)
    throw (nsga2::nsga2exception) ;
    individual& tournament(individual& ind1, individual& ind2) const;
    void crossover(const individual& parent1, const individual& parent2,
                   individual& child1, individual& child2);
    void realcross(const individual& parent1, const individual& parent2,
                   individual& child1, individual& child2);
    void bincross(const individual& parent1, const individual& parent2,
                   individual& child1, individual& child2);

    // Functions for logging
    void custom_report(population& pop);
    void init_streams();

    void report_parameters(std::ostream& os) const;
    void report_pop(const population& pop, std::ostream& os) const;

};

}

#endif /* NSGA2_H_ */
