#ifndef GLOBAL_H_
#define GLOBAL_H_

#include <vector>
#include <ostream>
#include <string>
#include <utility>

#include "nsga2/exception.h"
#include "nsga2/random.h"

#define EPS 1e-14
#define INF 1e+14

namespace nsga2 {

#define COMP_BETTER   0
#define COMP_WILCOXON 1
#define COMP_STUDENT  2

class NSGA2;
class population;

/// Configuration of NSGA-II individuals
struct individual_config {

    typedef void (*funcType)(double *,
                             double *,
                             int **,
                             double *,
                             std::vector<double> *,
                             double *);
    typedef void (*popFuncType)(population&);

    /// number of real-valued variables
    int nreal;
    int nbin;
    /// number of objectives
    int nobj;
    int nrep;
    int ncon;
    // int popsize;
    // int ngen;
    // double pcross_real;
    // double pcross_bin;
    double pmut_real;
    double pmut_bin;
    // double eta_c;
    double eta_m;
    /// Comparison method between individuals
    int comparison_method;
    std::vector<int> nbits;
    /// Search space of real-valued variables
    std::vector< std::pair<double,double> > limits_realvar;
    std::vector< std::pair<double,double> > limits_binvar;
    funcType function;
    double epsilon_c;
    double pvalue;
};

/// An NSGA-II individual
struct individual {

    individual() throw (); // needed for std::vector<individual> allocator

    /// Creates an individual according to the configuration given
    individual(const individual_config& c) throw (nsga2::nsga2exception);
    virtual ~individual();

    /// \brief Initialize an individual
    
    /// Initialize the position of the individual. When this function has been called,
    /// the individual has already one configuration. 
    void initialize() throw (nsga2::nsga2exception);

    void decode();
    /// \brief Evaluate an individual
    
    /// At the end of the evaluation, #obj will be filled with the resulting fitnesses
    void evaluate();
    std::pair<int,int> mutate();
    int real_mutate();
    int bin_mutate();

    /// \brief Check if an individual dominates the other
    /// \param b individual to be compared with
    /// \return  1 if this < b (this dominates b),
    ///          -1 if this > b (this is dominated by b),
    ///           0 if they are nondominated
    int check_dominance(const individual& b) const;

    int rank;
    double constr_violation;
    /// real-valued genes
    std::vector<double> xreal;
    std::vector< std::vector<int> > gene;
    std::vector<double> xbin;
    /// Fitnesses
    std::vector<double> obj;
    /// Specific for Meta-Optimization. If more complex metrics (e.g. non-parametrical tests)
    /// are used to compare individuals, you need fitnesses of all runs. This is the place to
    /// keep them. Not fully supported.
    std::vector<std::vector<double> > fullobj;
    std::vector<double> constr;
    /// Crowding distance
    double crowd_dist;

    int dcounter; // domination counter n_p
    std::vector<int> dominated;
    bool evaluated;

private:
    const individual_config* config;
    friend std::ostream& operator<< (std::ostream& os, const individual& ind);
    bool is_better(const individual b, int i) const;
};

std::ostream& operator<< (std::ostream& os, const individual& ind);

/// A Population of NSGA-II
class population {

public:
    population(const int size,
               const int nreal,
               const int nbin,
               const int nrep,
               const int ncon,
               const std::vector<int>& nbits,
               const std::vector< std::pair<double,double> >& limreal,
               const std::vector< std::pair<double,double> >& limbin,
               const int nobj,
               const double pmut_real,
               const double pmut_bin,
               const double eta_m,
               const double epsilon_c,
               const double pvalue,
               const int comparison_method,
               const individual_config::funcType func)
        throw (nsga2::nsga2exception);
    virtual ~population();

    void initialize() throw (nsga2::nsga2exception);

    void decode();
    /// Compute fitnesses for all individuals in the population
    void evaluate();
    void custom_evaluate(); // this one takes into account custom evaluation by user
    void set_popfunction(individual_config::popFuncType f);
    /// Non-dominated Sorting
    void fast_nds();
    /// Crowding distance on all fronts
    void crowding_distance_all();
    /// Crowding distance on a front
    /// \param: index of the front
    void crowding_distance(int fronti);

    std::pair<int,int> mutate();
    /// Merge two population
    void merge(const population& pop1, const population& pop2)
    throw (nsga2::nsga2exception);

    void report(std::ostream& os) const;
    void dump(std::ostream& os) const;
    void load(std::istream& os);

    int size() const {
        return ind.size();
    };

    /// \brief Set crowding distance type
    
    /// \param crowd if it is true, the crowding distance will be computed over objectives,
    /// otherwise on the parameters
    void set_crowdobj(bool crowd) {
        crowd_obj = crowd;
    };
    /// \brief Get crowding distance type
    
    /// If true, it means that the crowding distance is computed over objectives,
    /// otherwise on the parameters
    bool crowdobj() const { return crowd_obj; };

    std::vector<individual> ind;
    /// current generation
    int generation;

    /// List of fronts of this population. The index of the vector represents
    /// the rank
    std::vector< std::vector<int > > front;

private:

    individual_config ind_config;
    individual_config::popFuncType eval_pop_function;

    /// true: crowding over objective (default) false: crowding over real vars
    bool crowd_obj;

    void normal_evaluate();
    void normal_evaluate_openmp();

    friend std::ostream& operator<< (std::ostream& os, const population& pop);
};

std::ostream& operator<< (std::ostream& os, const population& pop);
}

#endif /* GLOBAL_H_ */
