/*
 * 
 * This test launches NSGA to optimize parameters of DE/PSO (according
 * to the first parameter passed to the executable). The second 
 * parameter passed is a configuration file containing the parameters
 * of NSGA and of the runs of DE/PSO. The results are saved into a 
 * folder with the experiment name (third parameter).
 * 
 * Now, NSGA can optimize several objectives:
 * 
 * - fitness: the value obtained at the end of the evolution
 * - eval: the number of evaluations needed to reach the FITNESS_TH
 * - gen: the number of generations needed to reach the FITNESS_TH
 * 
 * These metrics are then aggregated according using a selected
 * operation (such as mean, median, ecc.)
 * 
 */


#include <nsga2/global.h>
#include <nsga2/NSGA2.h>
#include <nsga2/NSGA2_DEPSO.h>
#include <depso/de.h>
#include <depso/pso.h>
#include <depso/de_crossover_strategies.h>
#include <depso/pso_best_functions.h>

#include "parametersParser.h"
#include "test_utilities.h"

#include <vector>
#include <map>
#include <iostream>
#include <sstream>
#include <utility>
#include <cmath>
#include <algorithm>

#define PI 3.14159265358979323846264338327950288419716939937510582097494

#define VERBOSE false

using namespace std;

// nsga parameters
int    NSGA_POPULATION_SIZE = -1;
int    NSGA_GENERATIONS     = -1;
double NSGA_CROSSOVER_RATE  = -1;
double NSGA_MUTATION_RATE   = -1;

// algorithm
string ALGORITHM = "NOT SET";

// problem (BOTH PSO AND DE)
int    PROBLEM_REPETITIONS = 1;                   // number of times the evolution process is repeated
int    NUM_GOALS = 2;                             // number of objectives for NSGA
int    MULTI_PROBLEM = 0;                         // if all goals relate to the same function (0) or to different ones (1)
double P_VALUE = -1;
int    COMPARISON_METHOD = 0;                     // how to say that an individual is better than another one
// Many goals
vector<string> GOALS;
// Many problems
vector<NsgaProblem> PROBLEMS;

void write_full_results(vector<double> & dst, vector<double> src)
{
    //for(int i=0;i<src.size();i++)
    //    cout << src[i] << " ";
    //cout << endl;
    dst = src;
}

void collect_results_multi(double* obj, vector<double>* fullobj, map< NsgaProblem, RESULTS_TYPE> results, vector<string> GOALS, vector<NsgaProblem> PROBLEMS)
{
    NsgaProblem problem;
	for(unsigned int goal=0; goal<GOALS.size(); goal++)
	{
        if(PROBLEMS.size() == GOALS.size()) // multi problem
            problem = PROBLEMS[goal];
        else
            problem = PROBLEMS[0];
        if(GOALS[goal] == "fitness_mean")
        {
            obj[goal] = mean(results[problem]["best_fitness"]);
            write_full_results(fullobj[goal], results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_logmean")
        {
            double tmp = log(mean(results[problem]["best_fitness"]));
            obj[goal] = tmp;
            write_full_results(fullobj[goal], results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_median")
        {
            obj[goal] = median(results[problem]["best_fitness"]);
            write_full_results(fullobj[goal], results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_min")
        {
            obj[goal] = min(results[problem]["best_fitness"]);
            write_full_results(fullobj[goal], results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_max")
        {
            obj[goal] = max(results[problem]["best_fitness"]);
            write_full_results(fullobj[goal], results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "eval_mean")
        {
            obj[goal] = mean(results[problem]["convergence_eval"]);
            write_full_results(fullobj[goal], results[problem]["convergence_eval"]);
        }
        else if(GOALS[goal] == "eval_median")
        {
            obj[goal] = median(results[problem]["convergence_eval"]);
            write_full_results(fullobj[goal], results[problem]["convergence_eval"]);
        }
        else if(GOALS[goal] == "eval_max")
        {
            obj[goal] = max(results[problem]["convergence_eval"]);
            write_full_results(fullobj[goal], results[problem]["convergence_eval"]);
        }
        else if(GOALS[goal] == "gen_mean")
        {
            obj[goal] = mean(results[problem]["convergence_gen"]);
            write_full_results(fullobj[goal], results[problem]["convergence_gen"]);
        }
        else if(GOALS[goal] == "gen_median")
        {
            obj[goal] = median(results[problem]["convergence_gen"]);
            write_full_results(fullobj[goal], results[problem]["convergence_gen"]);
        }
        else if(GOALS[goal] == "gen_max")
        {
            obj[goal] = max(results[problem]["convergence_gen"]);
            write_full_results(fullobj[goal], results[problem]["convergence_gen"]);
        }
        else
        {
            cout << "GOAL " << GOALS[goal] << " NOT VALID. Exiting!\n";
            exit(-1);
        }
	}
    
}

// convert an individual of NSGA-II to a set of valid DE parameters
void de_phenotype(double *particle, struct DECrossoverStrategyParameters & pars, int & pop_size, struct DECrossoverStrategy & cfunc)
{
	pars.CR = particle[DE_META_CR_INDEX] * (DE_META_CR_MAX - DE_META_CR_MIN) + DE_META_CR_MIN;
	pars.F = particle[DE_META_F_INDEX] * (DE_META_F_MAX - DE_META_F_MIN) + DE_META_F_MIN;
	pop_size = (int)(particle[DE_META_POPSIZE_INDEX] * (DE_META_POPSIZE_MAX - DE_META_POPSIZE_MIN) + DE_META_POPSIZE_MIN);
	const int mutation_function_id = argmax3(particle[DE_META_MUTATION_INDEX], particle[DE_META_MUTATION_INDEX+1], particle[DE_META_MUTATION_INDEX+2]);
	const int crossover_function_id = argmax2(particle[DE_META_CROSSOVER_INDEX], particle[DE_META_CROSSOVER_INDEX+1]);
	cfunc.func = de_crossover_strategy_functions[mutation_function_id][crossover_function_id];
	cfunc.data = &pars;
}

// convert an individual of NSGA-II to a set of valid PSO parameters
void pso_phenotype(double *particle, int & pop_size, double & w, double & c1, double & c2, struct PSOBestFunction & bfunc)
{
	pop_size = (int)(particle[PSO_META_POPSIZE_INDEX] * (PSO_META_POPSIZE_MAX - PSO_META_POPSIZE_MIN) + PSO_META_POPSIZE_MIN);
	w = particle[PSO_META_INERTIA_INDEX] * (PSO_META_INERTIA_MAX - PSO_META_INERTIA_MIN) + PSO_META_INERTIA_MIN;
	c1 = particle[PSO_META_CSELF_INDEX] * (PSO_META_CSELF_MAX - PSO_META_CSELF_MIN) + PSO_META_CSELF_MIN;
	c2 = particle[PSO_META_CGLOBAL_INDEX] * (PSO_META_CGLOBAL_MAX - PSO_META_CGLOBAL_MIN) + PSO_META_CGLOBAL_MIN;
   
	int topology_id = argmax3(particle[PSO_META_TOPOLOGY_INDEX], particle[PSO_META_TOPOLOGY_INDEX+1], particle[PSO_META_TOPOLOGY_INDEX+2]);
	
	bfunc.local = pso_local_best_functions[topology_id];
	bfunc.global = pso_global_best_functions[topology_id];
	
}

//TODO: handle goal
//TODO: move repetitions here to avoid multiple calls to X_phenotype
/// Runs a single run of the experiment
DEPSO_RESULTS run_problem(NsgaProblem p, double* params, string algorithm)
{
    // return value
    struct DEPSO_RESULTS results;
	// Parameters for DE, PSO
    // common
	int popsize;
    // DE only
    struct DECrossoverStrategyParameters pars;
    struct DECrossoverStrategy cfunc;
    // PSO only
    double w, c1, c2;
    struct PSOBestFunction bfunc;
    if(ALGORITHM == ALGO_DE)
    {
        de_phenotype(params, pars, popsize, cfunc);
        return DifferentialEvolution(
            popsize,          // population_size,
            p.func_num,       // fitness_function,
            &cfunc,           // strategy,
            p.generations,    // generation_count,
            p.evaluations,    // evaluations count
            p.stopping,       // use evals or gens
            p.fitness_th,     // fitness threshold
            p.dimension,      // problem_dimension,
            &p.bounds,        // boundaries,
            NULL,             // best_individual
            p.ss_improvement, // minimum improvement margin
            p.ss_gens         // max number of steady state generation
        );
    }
    else if (ALGORITHM == ALGO_PSO)
    {
		pso_phenotype(params, popsize, w, c1, c2, bfunc);
        return ParticleSwarmOptimization(
            popsize,          // population_size,
            w, c1, c2,        // confidence coefficients
            p.func_num,      // fitness_function,
            &bfunc,           // topology,
            p.generations,    // generation_count,
            p.evaluations,    // evaluations count
            p.stopping,       // use evals or gens
            p.fitness_th,     // fitness threshold
            p.dimension,      // param_count,
            &p.bounds,        // boundaries,
            NULL,             // best_individual,
            p.ss_improvement, // minimum improvement margin
            p.ss_gens         // max number of steady state generation
            );
    }
    else
    {
        cout << "Algorithm not valid!!! Exiting!" << endl;
        exit(1);
    }
}

void evaluate_ea_multi(double *xreal,
                       double *xbin,
                       int **gene,
                       double *obj,
                       vector<double> *fullobj,
                       double *constr)
{
    // Creates vectors with #PROBLEM_REPETITIONS elements to store results
	struct DEPSO_RESULTS results;
    map< NsgaProblem, RESULTS_TYPE> collected_results;
    // If single problem, PROBLEMS will contain only one element, otherwise N_GOALS
    for(int f=0;f<PROBLEMS.size();f++)
    {
        collected_results[PROBLEMS[f]]["best_fitness"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
        collected_results[PROBLEMS[f]]["convergence_eval"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
        collected_results[PROBLEMS[f]]["convergence_gen"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
    }
    // For each function to compute, call the correct algorithm
    for( map< NsgaProblem, RESULTS_TYPE>::iterator ii=collected_results.begin(); ii!=collected_results.end(); ++ii)
    {
        const NsgaProblem problem = (*ii).first;
        // Performs #PROBLEM_REPETITIONS of the selected EA and stores results
        for(int rep=0; rep < PROBLEM_REPETITIONS; rep++)
        {
            results = run_problem(problem, xreal, ALGORITHM);
            collected_results[(*ii).first]["best_fitness"][rep] = results.best_fitness;
            collected_results[(*ii).first]["convergence_eval"][rep] = results.convergence_evals;
            collected_results[(*ii).first]["convergence_gen"][rep] = results.convergence_gen;
        }
	}
	collect_results_multi(obj, fullobj, collected_results, GOALS, PROBLEMS);
}


int main(int argc, char *argv[]) {


	/*****************************************************/
	// check parameters
	/*****************************************************/
	if(argc<3)
	{
		cout << "Usage " << argv[0] << " DE|PSO PARAMETERS_FILE (EXPERIMENT_NAME)" << endl;
		exit(-1);
	}
	
	// set selected optimizer (DE,PSO)
	ALGORITHM = argv[1];
	if(ALGORITHM != "DE" && ALGORITHM != "PSO")
	{
		cout << "First parameter must be DE or PSO. Please check" << endl;
		exit(-1);
	}
	
	// read parameter file
	std::string parameters_file = argv[2]; 
	FILE *fp = fopen(parameters_file.c_str(), "r");
	if(fp==NULL)
	{
		cout << "Could not open " << parameters_file << ". Please Check!\n";
		exit(-1);
	}
	fclose(fp);
	
	parametersParser pp;
	int tmp;

	// nsga parameters
	pp.set_parameter(parameters_file.c_str(), "NSGA_POPULATION_SIZE", &NSGA_POPULATION_SIZE, true);
	pp.set_parameter(parameters_file.c_str(), "NSGA_GENERATIONS", &NSGA_GENERATIONS, true);
	pp.set_parameter(parameters_file.c_str(), "NSGA_CROSSOVER_RATE", &NSGA_CROSSOVER_RATE, true);
	pp.set_parameter(parameters_file.c_str(), "NSGA_MUTATION_RATE", &NSGA_MUTATION_RATE, true);
    pp.set_parameter(parameters_file.c_str(), "P_VALUE", &P_VALUE);
    pp.set_parameter(parameters_file.c_str(), "PROBLEM_REPETITIONS", &PROBLEM_REPETITIONS, true);
    pp.set_parameter(parameters_file.c_str(), "COMPARISON_METHOD", &COMPARISON_METHOD, true);
    // general problem options
    pp.set_parameter(parameters_file.c_str(), "NUM_GOALS", &NUM_GOALS, true);
    pp.set_parameter(parameters_file.c_str(), "MULTI_PROBLEM", &MULTI_PROBLEM, true);
    PROBLEMS.resize(0);
    GOALS.resize(NUM_GOALS);
    stringstream ss;
    // now collect goals
    for(int g=0; g<NUM_GOALS; g++)
    {
		ss.str("");
		ss << "GOAL_" << g;
		pp.set_parameter(parameters_file.c_str(), ss.str().c_str(), &(GOALS[g]), true);
    }
    // single problem
    if(!MULTI_PROBLEM)
	{
        NsgaProblem single_problem;
        single_problem = read_problem_from_file(parameters_file);
        single_problem.set_stopping_criterion(GOALS);
        PROBLEMS.push_back(single_problem);
	}
    // get everything if multi problem
    else
    {
        for(int g=0; g<NUM_GOALS; g++)
        {
            NsgaProblem problem_tmp;
            problem_tmp = read_problem_from_file(parameters_file, g);
            problem_tmp.set_stopping_criterion(GOALS[g]);
            PROBLEMS.push_back(NsgaProblem(problem_tmp));
		}
	}
	
	// set experiment name
	std::string experiment_name = static_cast<ostringstream*>( &(ostringstream() << time(0)) )->str();
	if(argc>=4)
		experiment_name = argv[3];
	
    nsga2::individual_config conf;
    
    int PARTICLE_SIZE = ALGORITHM == "DE" ? DE_META_PARTICLE_SIZE : PSO_META_PARTICLE_SIZE;
    
	for(int i=0;i<PARTICLE_SIZE;i++)
		conf.limits_realvar.push_back(make_pair(0.0,1.0));
    
    // compute an hash on experiment name, so if two experiments starts at the same time
    // they will have different random seed
    int hash = 0;
	int offset = 'a' - 1;
	for(string::const_iterator it=experiment_name.begin(); it!=experiment_name.end(); ++it) {
	  hash = hash << 1 ^ (*it - offset);
	}
    
    int seed = time(0) + hash;
    cout << "Using seed " << seed << endl;

    nsga2::NSGA2 nsga2;
    nsga2.set_experiment_name(experiment_name);
    nsga2.set_seed(seed);
    nsga2.set_nreal(PARTICLE_SIZE);
    nsga2.set_nbin(0);
    nsga2.set_nrep(PROBLEM_REPETITIONS);
    nsga2.set_pvalue(P_VALUE);
    nsga2.set_comparison(COMPARISON_METHOD);
    nsga2.set_nobj(NUM_GOALS); // number of objectives
    nsga2.set_ncon(0); // add a constraint due to possible simulation failures
    nsga2.set_popsize(NSGA_POPULATION_SIZE);
    nsga2.set_ngen(NSGA_GENERATIONS);
    nsga2.set_pcross_real(NSGA_CROSSOVER_RATE);
    nsga2.set_pcross_bin(0.0);
    nsga2.set_pmut_real(NSGA_MUTATION_RATE);
    nsga2.set_pmut_bin(0.0);
    nsga2.set_eta_c(10);
    nsga2.set_eta_m(10);
    nsga2.set_epsilon_c(1e-14);
    //nsga2.set_nbits(0);
    nsga2.set_limits_realvar(conf.limits_realvar);
    // nsga2.set_limits_binvar(limits_binvar);
    nsga2.set_function(&evaluate_ea_multi);
    nsga2.set_crowdobj(false); // crowd over the parameters, not the objective functions
	// nsga2.set_crowdobj(true); // crowd over objective function
    // nsga2.set_custom_report_function(&update_generation);
    // nsga2.set_nreport(10);
    nsga2.set_backup_filename(""); // no backup

    /**** NSGA RUN ****/
    nsga2.initialize();
    

    // Write information about the problem
    std::ofstream fpt1;
	ostringstream filename;
	filename.str("");
	filename << nsga2.get_result_dir() << "/nsga2_optimizer_parameters.out";
    fpt1.open(filename.str().c_str(), ios::out | ios::trunc);
    fpt1 << "ALGORITHM "      << ALGORITHM << endl;
    fpt1 << "NUM_GOALS "      << NUM_GOALS << endl;
    for(int g=0; g<NUM_GOALS; g++)
    {
        fpt1 << "GOAL_" << g << " " << GOALS[g] << endl;
    }
    fpt1 << "MULTI_FUNCTION " << MULTI_PROBLEM << endl;
    if(!MULTI_PROBLEM)
    {
        fpt1 << "SINGLE PROBLEM" << endl;
        fpt1 << PROBLEMS[0];
    }
    else
    {
        for(int g=0; g<NUM_GOALS; g++)
        {
            fpt1 << "PROBLEM " << g << endl;
            fpt1 << PROBLEMS[g];
        }
    }
	fpt1.flush();
	fpt1.close();
	
    nsga2.evolve();
    
    cout << "COMPLETED\n";
    
    return 0;
}
