/*
 * 
 * This test launches a single evaluation of DE/PSO, as specified via
 * command line. The results are shown to stdout.
 * 
 * The parameters can be passed by command line one by one or using a
 * configuration file just like the ones used by test_multi.
 * 
 */

#include <nsga2/global.h>
#include <nsga2/NSGA2.h>
#include <nsga2/NSGA2_DEPSO.h>
#include <depso/de.h>
#include <depso/pso.h>
#include <depso/test_func.h>
#include <depso/de_crossover_strategies.h>
#include <depso/pso_best_functions.h>
#include <depso/random.h>

#include "helper_strings.h"
#include "test_utilities.h"
#include "parametersParser.h"

#include <vector>
#include <iostream>
#include <sstream>
#include <utility>
#include <cmath>
#include <algorithm>

#define PI 3.14159265358979323846264338327950288419716939937510582097494

using namespace std;

void collect_results(double* obj, map< NsgaProblem, RESULTS_TYPE> results, vector<string> GOALS, vector<NsgaProblem> PROBLEMS, int verbose)
{
    NsgaProblem problem;
	for(unsigned int goal=0; goal<GOALS.size(); goal++)
	{
        if(PROBLEMS.size() == GOALS.size()) // multi problem
            problem = PROBLEMS[goal];
        else
            problem = PROBLEMS[0];
        if(GOALS[goal] == "fitness_mean")
        {
            obj[goal] = mean(results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_logmean")
        {
            double tmp = log(mean(results[problem]["best_fitness"]));
            obj[goal] = tmp;
        }
        else if(GOALS[goal] == "fitness_median")
        {
            obj[goal] = median(results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_min")
        {
            obj[goal] = min(results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_max")
        {
            obj[goal] = max(results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "eval_mean")
        {
            obj[goal] = mean(results[problem]["convergence_eval"]);
        }
        else if(GOALS[goal] == "eval_median")
        {
            obj[goal] = median(results[problem]["convergence_eval"]);
        }
        else if(GOALS[goal] == "eval_max")
        {
            obj[goal] = max(results[problem]["convergence_eval"]);
        }
        else if(GOALS[goal] == "gen_mean")
        {
            obj[goal] = mean(results[problem]["convergence_gen"]);
        }
        else if(GOALS[goal] == "gen_median")
        {
            obj[goal] = median(results[problem]["convergence_gen"]);
        }
        else if(GOALS[goal] == "gen_max")
        {
            obj[goal] = max(results[problem]["convergence_gen"]);
        }
        else
        {
            cout << "GOAL " << GOALS[goal] << " NOT VALID. Exiting!\n";
            exit(-1);
        }
    if(verbose)
		cout << GOALS[goal] << " - " << problem.func_name << " (" << problem.dimension << ") : " << obj[goal] << endl;
	else
		cout << obj[goal] << endl;
	}
    
}

int main(int argc, char *argv[]) {

	/*****************************************************/
	// READ ALL PARAMETERS
	/*****************************************************/
	if(argc<2)
	{
		cout << "You need to specify at least DE or PSO" << endl;
		exit(-1);
	}
	
	// set selected optimizer (DE,PSO)
	std::string optimizer_str = argv[1];
	if(optimizer_str != "DE" && optimizer_str != "PSO")
	{
		cout << "First parameter must be DE or PSO. Please check" << endl;
		exit(-1);
	}
	
    // Many goals
    int NUM_GOALS = -1;
    vector<string> GOALS;
    // Many problems
    int MULTI_PROBLEM = -1;
    vector<NsgaProblem> PROBLEMS;
    // Many repetitions
    int PROBLEM_REPETITIONS = -1;

    // Init random
    alea_seed(time(0));
    
    // DE default values
    float F = 0.9;
    float CR = 0.9;
    int mutation_id = 0;
    int crossover_id = 0;
    struct DECrossoverStrategy cfunc;
    struct DECrossoverStrategyParameters pars;
    
    // PSO default values
    float C1 = 1.19;
    float C2 = 1.19;
    float W = 0.7;
    int topology_id = 0;
    struct PSOBestFunction bfunc;
    
    // Common default values
    int verbose = 1;
    int pop_size = 64;
    int problem_dimension = 10;
    int eval_num = -1;
    int gen_num = -1;
    int runs_num = 1;
    StoppingCriterion stopping_criterion = STOP_EVALS; //evaluations
    float fitness_goal = 0.0;
    string fitness_name = "griewank";
    string experiment_name = "experiment";
    string configuration_file = "";
    //TODO: output to file

	/* verbosity */
	if( checkCmdLineFlag(argc, (const char**)argv, "v") )
	{
		verbose = getCmdLineArgumentInt (argc, (const char **)argv, "v");
		if(verbose > 0)
            cout << "Verbosity set to " << verbose << endl;
        if(verbose < 0)
        {
            cout << "Verbosity is outside boundaries. Exiting." << endl;
            exit(-1);
        }
	}    
    // Now read all parameters. First, the one related to DE or PSO
    if(optimizer_str == "DE")
    {
		/* CR */
		if (checkCmdLineFlag (argc, (const char **) argv, "CR"))
		{
			CR = getCmdLineArgumentFloat (argc, (const char **)argv, "CR");
			if ((CR >= DE_META_CR_MIN) && (CR <= DE_META_CR_MAX))
			{
				if(verbose)	cout << "Value of CR set to " << CR << endl;
			}
            else
            {
                cout << "Value of CR is outside boundaries. Exiting." << endl;
                exit(-1);
            }
		}
		/* F */
		if (checkCmdLineFlag (argc, (const char **) argv, "F"))
		{
			F = getCmdLineArgumentFloat (argc, (const char **)argv, "F");
			if ((F >= DE_META_F_MIN) && (F <= DE_META_F_MAX))
			{
				if(verbose)	cout << "Value of F set to " << F << endl;
			}
            else
            {
                cout << "Value of F is outside boundaries. Exiting." << endl;
                exit(-1);
            }
		}
        pars.CR = CR;
        pars.F = F;
        /* mutation */
		if (checkCmdLineFlag (argc, (const char**)argv, "m"))
		{
			mutation_id = getCmdLineArgumentInt (argc, (const char**)argv, "m");
			if ((mutation_id >= 0) && (mutation_id <= 2))
			{
				if(verbose)	cout << "Mutation id set to " << mutation_id << endl;
			}
            else
            {
                cout << "Mutation id is outside boundaries. Exiting." << endl;
                exit(-1);
            }
		}
		/* crossover */
		if (checkCmdLineFlag (argc, (const char**)argv, "x"))
		{
			crossover_id = getCmdLineArgumentInt (argc, (const char**)argv, "x");
			if ((crossover_id >= 0) && (crossover_id <= 1))
			{
				if(verbose)	cout << "Crossover id set to " << crossover_id << endl;
			}
            else
            {
                cout << "Crossover id is outside boundaries. Exiting." << endl;
                exit(-1);
            }
		}
        cfunc.func = de_crossover_strategy_functions[mutation_id][crossover_id];
        cfunc.data = &pars;
    }
    else //PSO
    {
		/* C1 */
		if (checkCmdLineFlag (argc, (const char **) argv, "C1"))
		{
			C1 = getCmdLineArgumentFloat (argc, (const char **)argv, "C1");
			if ((C1 >= PSO_META_CSELF_MIN) && (C1 <= PSO_META_CSELF_MAX))
			{
				if(verbose)	cout << "Value of C1 set to " << C1 << endl;
			}
            else
            {
                cout << "Value of C1 is outside boundaries. Exiting." << endl;
                exit(-1);
            }
		}
		/* C2 */
		if (checkCmdLineFlag (argc, (const char **) argv, "C2"))
		{
			C2 = getCmdLineArgumentFloat (argc, (const char **)argv, "C2");
			if ((C2 >= PSO_META_CGLOBAL_MIN) && (C2 <= PSO_META_CGLOBAL_MAX))
			{
				if(verbose)	cout << "Value of C2 set to " << C2 << endl;
			}
            else
            {
                cout << "Value of C2 is outside boundaries. Exiting." << endl;
                exit(-1);
            }
		}
		/* Inertia */
		if (checkCmdLineFlag (argc, (const char **) argv, "W"))
		{
			W = getCmdLineArgumentFloat (argc, (const char **)argv, "W");
			if ((W >= PSO_META_INERTIA_MIN) && (W <= PSO_META_INERTIA_MAX))
			{
				if(verbose)	cout << "Value of Inertia set to " << W << endl;
			}
            else
            {
                cout << "Value of Inertia is outside boundaries. Exiting." << endl;
                exit(-1);
            }
		}
		/* Topology */
		if (checkCmdLineFlag (argc, (const char **) argv, "T"))
		{
			topology_id = getCmdLineArgumentInt (argc, (const char **)argv, "T");
			if ((topology_id >= 0) && (topology_id <= 2))
			{
				if(verbose)	cout << "Topology ID set to " << topology_id << endl;
			}
            else
            {
                cout << "Topology ID is outside boundaries. Exiting." << endl;
                exit(-1);
            }
		}
        bfunc.local = pso_local_best_functions[topology_id];
        bfunc.global = pso_global_best_functions[topology_id];
    }
    // Then things in common between DE and PSO
	/* population size */
	if( checkCmdLineFlag(argc, (const char**)argv, "p") )
	{
		pop_size = getCmdLineArgumentInt (argc, (const char **)argv, "p");
		if (pop_size > 1)
		{
			if(verbose)	cout << "Population size set to " << pop_size << endl;
		}
        else
        {
            cout << "Population size is outside boundaries. Exiting." << endl;
            exit(-1);
        }
	}
	/* number of runs */
	if( checkCmdLineFlag(argc, (const char**)argv, "r") )
	{
		runs_num = getCmdLineArgumentInt (argc, (const char **)argv, "r");
		if (runs_num >= 1)
		{
			if(verbose)	cout << "Runs number set to " << runs_num << endl;
		}
        else
        {
            cout << "Runs number is outside boundaries. Exiting." << endl;
            exit(-1);
        }
	}
    // You can use a configuration file, or select everything manually
    if (checkCmdLineFlag (argc, (const char**)argv, "conf"))
    {
        configuration_file = getCmdLineArgumentString(argc, (const char**)argv, "conf");
        FILE *fp = fopen(configuration_file.c_str(), "r");
        if(fp==NULL)
        {
            cout << "Could not open " << configuration_file << ". Please Check!\n";
            exit(-1);
        }
        fclose(fp);
        // ok, file is valid, now read what to do, similarly to test_nsga_multigoal
        // general problem options
        parametersParser pp;
        pp.set_parameter(configuration_file.c_str(), "PROBLEM_REPETITIONS", &PROBLEM_REPETITIONS, true);
        pp.set_parameter(configuration_file.c_str(), "NUM_GOALS", &NUM_GOALS, true);
        pp.set_parameter(configuration_file.c_str(), "MULTI_PROBLEM", &MULTI_PROBLEM, true);
        PROBLEMS.resize(0);
        GOALS.resize(NUM_GOALS);
        stringstream ss;
        // now collect goals
        for(int g=0; g<NUM_GOALS; g++)
        {
            ss.str("");
            ss << "GOAL_" << g;
            pp.set_parameter(configuration_file.c_str(), ss.str().c_str(), &(GOALS[g]), true);
        }
        // single problem
        if(!MULTI_PROBLEM)
        {
            NsgaProblem single_problem;
            single_problem = read_problem_from_file(configuration_file);
            single_problem.set_stopping_criterion(GOALS);
            PROBLEMS.push_back(single_problem);
        }
        // get stuff if multi problem
        else
        {
            for(int g=0; g<NUM_GOALS; g++)
            {
                NsgaProblem problem_tmp;
                problem_tmp = read_problem_from_file(configuration_file, g);
                problem_tmp.set_stopping_criterion(GOALS[g]);
                PROBLEMS.push_back(NsgaProblem(problem_tmp));
            }
        }
    }
    else // no configuration file, read from here (TODO: old style, only fitness vs evals)
    {
        NsgaProblem problem_tmp;
        /* dimension of the problem */
        if (checkCmdLineFlag (argc, (const char**)argv, "d"))
        {
            problem_dimension = getCmdLineArgumentInt (argc, (const char**)argv, "d");
            if (problem_dimension > 1)
			{
				if(verbose)	cout << "Problem dimension set to " << problem_dimension << endl;
			}
            else
            {
                cout << "Problem dimension is outside boundaries. Exiting." << endl;
                exit(-1);
            }
        }
        problem_tmp.dimension = problem_dimension;
        /* number of generations */
        if( checkCmdLineFlag(argc, (const char**)argv, "g") )
        {
            gen_num = getCmdLineArgumentInt (argc, (const char **)argv, "g");
            if (gen_num > 1)
			{
				if(verbose)	cout << "Generation number set to " << gen_num << endl;
			}
            else
            {
                cout << "Generation number is outside boundaries. Exiting." << endl;
                exit(-1);
            }
        }
        problem_tmp.generations = gen_num;
        /* number of evaluation */
        if( checkCmdLineFlag(argc, (const char**)argv, "e") )
        {
            eval_num = getCmdLineArgumentInt (argc, (const char **)argv, "e");
            if (eval_num > 1)
			{
				if(verbose)	cout << "Evaluation number set to " << eval_num << endl;
			}
            else
            {
                cout << "Evaluation number is outside boundaries. Exiting." << endl;
                exit(-1);
            }
        }
        problem_tmp.evaluations = eval_num;
        /* fitness goal */
        if( checkCmdLineFlag(argc, (const char**)argv, "goal") )
        {
            fitness_goal = getCmdLineArgumentFloat (argc, (const char **)argv, "goal");
            if (fitness_goal >= 1e-20)
			{
				if(verbose)	cout << "Fitness goal set to " << fitness_goal << endl;
			}
            else
            {
                cout << "Fitness goal is outside boundaries. Exiting." << endl;
                exit(-1);
            }
        }
        problem_tmp.fitness_th = fitness_goal;
        /* fitness function */
        if( checkCmdLineFlag(argc, (const char**)argv, "fun") )
        {
            fitness_name = getCmdLineArgumentString (argc, (const char **)argv, "fun");
            if(verbose)
				cout << "Fitness name set to " << fitness_name << endl;
        }
        problem_tmp.func_name = fitness_name;
        problem_tmp.func_num = get_function_id_by_name(fitness_name);
        double min_bound, max_bound;
        boundaries_by_function_name(fitness_name, min_bound, max_bound);
        DEInitBoundaries(&(problem_tmp.bounds), problem_tmp.dimension, min_bound, max_bound);
        // TODO: handle goals
        GOALS.push_back("fitness_mean");
        // TODO: handle repetitions and runs
        PROBLEM_REPETITIONS = 1;

        problem_tmp.set_stopping_criterion(GOALS);
        PROBLEMS.push_back(problem_tmp);
    }
    if(verbose)
    {
		cout << "*******************************\n";
		cout << "**** Starting optimization ****\n";
		cout << "*******************************\n";
    
		cout << "GOALS\n\n";
		for(int g=0; g<GOALS.size(); g++)
			cout << GOALS[g] << endl;
		cout << endl;
		cout << "PROBLEMS\n\n";
		for(int g=0; g<PROBLEMS.size(); g++)
			cout << PROBLEMS[g] << endl;
    }


    for(int run=0; run < runs_num; run++)
    {
        struct DEPSO_RESULTS results;
        map< NsgaProblem, RESULTS_TYPE> collected_results;
        // If single problem, PROBLEMS will contain only one element, otherwise N_GOALS
        for(int f=0;f<PROBLEMS.size();f++)
        {
            collected_results[PROBLEMS[f]]["best_fitness"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
            collected_results[PROBLEMS[f]]["convergence_eval"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
            collected_results[PROBLEMS[f]]["convergence_gen"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
        }
        // For each function to compute, call the correct algorithm
        for( map< NsgaProblem, RESULTS_TYPE>::iterator ii=collected_results.begin(); ii!=collected_results.end(); ++ii)
        {
            NsgaProblem p = (*ii).first;
            // Performs #PROBLEM_REPETITIONS of the selected EA and stores results
            for(int rep=0; rep < PROBLEM_REPETITIONS; rep++)
            {
                if(optimizer_str == ALGO_DE)
                {
                    results = DifferentialEvolution(
                        pop_size,         // population_size,
                        p.func_num,       // fitness_function,
                        &cfunc,           // strategy,
                        p.generations,    // generation_count,
                        p.evaluations,    // evaluations count
                        p.stopping,       // use evals or gens
                        p.fitness_th,     // fitness threshold
                        p.dimension,      // problem_dimension,
                        &p.bounds,        // boundaries,
                        NULL,             // best_individual
                        p.ss_improvement, // minimum improvement margin
                        p.ss_gens         // max number of steady state generation
                    );
                }
                else if (optimizer_str == ALGO_PSO)
                {
                    results = ParticleSwarmOptimization(
                        pop_size,          // population_size,
                        W, C1, C2,        // confidence coefficients
                        p.func_num,       // fitness_function,
                        &bfunc,           // topology,
                        p.generations,    // generation_count,
                        p.evaluations,    // evaluations count
                        p.stopping,       // use evals or gens
                        p.fitness_th,     // fitness threshold
                        p.dimension,      // param_count,
                        &p.bounds,        // boundaries,
                        NULL,             // best_individual,
                        p.ss_improvement, // minimum improvement margin
                        p.ss_gens         // max number of steady state generation
                        );
                }
                else
                {
                    cout << "Algorithm not valid!!! Exiting!" << endl;
                    exit(1);
                }
                
                collected_results[(*ii).first]["best_fitness"][rep] = results.best_fitness;
                collected_results[(*ii).first]["convergence_eval"][rep] = results.convergence_evals;
                collected_results[(*ii).first]["convergence_gen"][rep] = results.convergence_gen;
            }
        }
        double obj[NUM_GOALS];
        collect_results(obj, collected_results, GOALS, PROBLEMS, verbose);
    }

    return 0;
}
