/*
 * 
 * This test launches a systematic evaluation of the parameters of DE/PSO,
 * as specified via command line. The results are saved to file (if provided)
 * or printed to file.
 * 
 * The parameters can be passed by command line one by one or using a
 * configuration file just like the ones used by test_multi (last not
 * implemented yet).
 * 
 */
#include <nsga2/NSGA2_DEPSO.h>
#include <depso/random.h>
#include <depso/de.h>
#include <depso/pso.h>
#include <depso/de_crossover_strategies.h>
#include <depso/pso_best_functions.h>

#include "helper_strings.h"
#include "test_utilities.h"
#include "parametersParser.h"

#include <vector>
#include <iostream>
#include <sstream>
#include <utility>
#include <stdlib.h>
#include <algorithm>

using namespace std;

// TODO pass by command line?
#define DE_SYS_F_STEP 0.2
#define DE_SYS_CR_STEP 0.1
#define DE_SYS_POPSIZE_STEP 8
#define PSO_SYS_CSELF_STEP 0.5
#define PSO_SYS_CGLOBAL_STEP 0.5
#define PSO_SYS_INERTIA_STEP 0.5
#define PSO_SYS_POPSIZE_STEP 16

void collect_results(double* obj, map< NsgaProblem, RESULTS_TYPE> results, vector<string> GOALS, vector<NsgaProblem> PROBLEMS, ostream &f)
{
    NsgaProblem problem;
	for(unsigned int goal=0; goal<GOALS.size(); goal++)
	{
        if(PROBLEMS.size() == GOALS.size()) // multi problem
            problem = PROBLEMS[goal];
        else
            problem = PROBLEMS[0];
        if(GOALS[goal] == "fitness_mean")
        {
            obj[goal] = mean(results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_logmean")
        {
            double tmp = log(mean(results[problem]["best_fitness"]));
            obj[goal] = tmp;
        }
        else if(GOALS[goal] == "fitness_median")
        {
            obj[goal] = median(results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_min")
        {
            obj[goal] = min(results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "fitness_max")
        {
            obj[goal] = max(results[problem]["best_fitness"]);
        }
        else if(GOALS[goal] == "eval_mean")
        {
            obj[goal] = mean(results[problem]["convergence_eval"]);
        }
        else if(GOALS[goal] == "eval_median")
        {
            obj[goal] = median(results[problem]["convergence_eval"]);
        }
        else if(GOALS[goal] == "eval_max")
        {
            obj[goal] = max(results[problem]["convergence_eval"]);
        }
        else if(GOALS[goal] == "gen_mean")
        {
            obj[goal] = mean(results[problem]["convergence_gen"]);
        }
        else if(GOALS[goal] == "gen_median")
        {
            obj[goal] = median(results[problem]["convergence_gen"]);
        }
        else if(GOALS[goal] == "gen_max")
        {
            obj[goal] = max(results[problem]["convergence_gen"]);
        }
        else
        {
            cout << "GOAL " << GOALS[goal] << " NOT VALID. Exiting!\n";
            exit(-1);
        }
        if(f)
            f << " " << obj[goal];
	}    
}

int main(int argc, char *argv[]) {

    // Many goals
    int NUM_GOALS = -1;
    vector<string> GOALS;
    // Many problems
    int MULTI_PROBLEM = -1;
    vector<NsgaProblem> PROBLEMS;
    // Many repetitions
    int PROBLEM_REPETITIONS = -1;
    // Configuration file (set from command line)
    string configuration_file = "";

	/*****************************************************/
	// READ ALL PARAMETERS
	/*****************************************************/
	if(argc<2)
	{
		cout << "You need to specify at least DE or PSO" << endl;
		exit(-1);
	}
	
	// set selected optimizer (DE,PSO)
	std::string optimizer_str = argv[1];
	if(optimizer_str != "DE" && optimizer_str != "PSO")
	{
		cout << "First parameter must be DE or PSO. Please check" << endl;
		exit(-1);
	}
        
    // Init random
    alea_seed(time(0));

    // some params

	/* verbosity */
    int verbose = 1;
	if( checkCmdLineFlag(argc, (const char**)argv, "v") )
	{
		verbose = getCmdLineArgumentInt (argc, (const char **)argv, "v");
		if(verbose > 0)
            cout << "Verbosity set to " << verbose << endl;
        if(verbose < 0)
        {
            cout << "Verbosity is outside boundaries. Exiting." << endl;
            exit(-1);
        }
	}
    bool to_outstd = true;
    string outfile = "";
    ofstream outstr;
    if( checkCmdLineFlag(argc, (const char**)argv, "out") )
    {
        outfile = getCmdLineArgumentString (argc, (const char **)argv, "out");
        outstr.open(outfile.c_str());
        if(outstr.is_open())
        {
            to_outstd = false;
        }
        else
        {
            cout << "Cannot open output file: " << outfile << endl;
            exit(1);
        }
    }
    streambuf * buf;
    if(to_outstd)
        buf = cout.rdbuf();
    else
        buf = outstr.rdbuf();
    ostream out(buf);

    // Read problems from a configuration file
    // TODO: try to keep it similar to the other files
    if (checkCmdLineFlag (argc, (const char**)argv, "conf"))
    {
        configuration_file = getCmdLineArgumentString(argc, (const char**)argv, "conf");
        FILE *fp = fopen(configuration_file.c_str(), "r");
        if(fp==NULL)
        {
            cout << "Could not open " << configuration_file << ". Please Check!\n";
            exit(-1);
        }
        fclose(fp);
        // ok, file is valid, now read what to do, similarly to test_nsga_multigoal
        // general problem options
        parametersParser pp;
        pp.set_parameter(configuration_file.c_str(), "PROBLEM_REPETITIONS", &PROBLEM_REPETITIONS, true);
        pp.set_parameter(configuration_file.c_str(), "NUM_GOALS", &NUM_GOALS, true);
        pp.set_parameter(configuration_file.c_str(), "MULTI_PROBLEM", &MULTI_PROBLEM, true);
        PROBLEMS.resize(0);
        GOALS.resize(NUM_GOALS);
        stringstream ss;
        // now collect goals
        for(int g=0; g<NUM_GOALS; g++)
        {
            ss.str("");
            ss << "GOAL_" << g;
            pp.set_parameter(configuration_file.c_str(), ss.str().c_str(), &(GOALS[g]), true);
        }
        // single problem
        if(!MULTI_PROBLEM)
        {
            NsgaProblem single_problem;
            single_problem = read_problem_from_file(configuration_file);
            single_problem.set_stopping_criterion(GOALS);
            PROBLEMS.push_back(single_problem);
        }
        // get stuff if multi problem
        else
        {
            for(int g=0; g<NUM_GOALS; g++)
            {
                NsgaProblem problem_tmp;
                problem_tmp = read_problem_from_file(configuration_file, g);
                problem_tmp.set_stopping_criterion(GOALS[g]);
                PROBLEMS.push_back(NsgaProblem(problem_tmp));
            }
        }
    }
    else // no configuration file, read from here (TODO: old style, only fitness vs evals)
    {
        cout << "You need to pass a configuration file. Sorry about that...\n";
        exit(1);
    }

    if(verbose)
    {
		cout << "*******************************\n";
		cout << "**** Starting optimization ****\n";
		cout << "*******************************\n";
    }
    out << "GOALS\n\n";
    for(int g=0; g<GOALS.size(); g++)
        out << GOALS[g] << endl;
    out << endl;
    out << "PROBLEMS\n\n";
    for(int g=0; g<PROBLEMS.size(); g++)
        out << PROBLEMS[g] << endl;

    // For collecting results
    struct DEPSO_RESULTS results;
    map< NsgaProblem, RESULTS_TYPE> collected_results;
    double obj[NUM_GOALS];
    // Optimize it!
    if(optimizer_str == "DE")
    {
        // Write header to file (or stdout)
        out << "START OPTIMIZATION DE\n";
        out << "POPSIZE F CR MUTATION CROSSOVER";
        for(int i=0;i<NUM_GOALS;i++)
            out << " GOAL_" << i;
        out << endl;
        
        // DE parameters
        struct DECrossoverStrategy cfunc;
        struct DECrossoverStrategyParameters pars;
        // If single problem, PROBLEMS will contain only one element, otherwise N_GOALS
        for(int f=0;f<PROBLEMS.size();f++)
        {
            collected_results[PROBLEMS[f]]["best_fitness"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
            collected_results[PROBLEMS[f]]["convergence_eval"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
            collected_results[PROBLEMS[f]]["convergence_gen"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
        }

        for(int pop_size=DE_META_POPSIZE_MIN; pop_size<=DE_META_POPSIZE_MAX; pop_size+=DE_SYS_POPSIZE_STEP)
            for(double F=DE_META_F_MIN; F<=DE_META_F_MAX; F+=DE_SYS_F_STEP)
                for(double CR=DE_META_CR_MIN; CR<=DE_META_CR_MAX; CR+=DE_SYS_CR_STEP)
                    for(int mut_type=0; mut_type<DE_META_MUTATION_COUNT; mut_type+=1)
                        for(int cross_type=0; cross_type<DE_META_CROSSOVER_COUNT; cross_type+=1)
                        {
                            pars.CR = CR;
                            pars.F = F;
                            cfunc.func = de_crossover_strategy_functions[mut_type][cross_type];
                            cfunc.data = &pars;
                            if(verbose && !to_outstd)
                                cout << "Testing " << pop_size << " " << F << " " << CR << " " << mut_type << " " << cross_type << endl; 
                            out << pop_size << " " << F << " " << CR << " " << mut_type << " " << cross_type; 
                            // For each function to compute, call the correct algorithm
                            for( map< NsgaProblem, RESULTS_TYPE>::iterator ii=collected_results.begin(); ii!=collected_results.end(); ++ii)
                            {
                                NsgaProblem p = (*ii).first;
                                // Performs #PROBLEM_REPETITIONS of the selected EA and stores results
                                for(int rep=0; rep < PROBLEM_REPETITIONS; rep++)
                                {
                                    results = DifferentialEvolution(
                                        pop_size,         // population_size,
                                        p.func_num,       // fitness_function,
                                        &cfunc,           // strategy,
                                        p.generations,    // generation_count,
                                        p.evaluations,    // evaluations count
                                        p.stopping,       // use evals or gens
                                        p.fitness_th,     // fitness threshold
                                        p.dimension,      // problem_dimension,
                                        &p.bounds,        // boundaries,
                                        NULL,             // best_individual
                                        p.ss_improvement, // minimum improvement margin
                                        p.ss_gens         // max number of steady state generation
                                    );
                                    collected_results[(*ii).first]["best_fitness"][rep] = results.best_fitness;
                                    collected_results[(*ii).first]["convergence_eval"][rep] = results.convergence_evals;
                                    collected_results[(*ii).first]["convergence_gen"][rep] = results.convergence_gen;
                                }
                            }
                            collect_results(obj, collected_results, GOALS, PROBLEMS, out);
                            out << endl;
                        }
    }
    else // PSO
    {
        // Write header to file (or stdout)
        out << "START OPTIMIZATION PSO\n";
        out << "POPSIZE CSELF CGLOB W TOPOLOGY";
        for(int i=0;i<NUM_GOALS;i++)
            out << " GOAL_" << i;
        out << endl;
        
        // PSO parameters
        struct PSOBestFunction bfunc;
        // If single problem, PROBLEMS will contain only one element, otherwise N_GOALS
        for(int f=0;f<PROBLEMS.size();f++)
        {
            collected_results[PROBLEMS[f]]["best_fitness"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
            collected_results[PROBLEMS[f]]["convergence_eval"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
            collected_results[PROBLEMS[f]]["convergence_gen"] = vector<double>(PROBLEM_REPETITIONS, 0.0);
        }

        for(int pop_size=PSO_META_POPSIZE_MIN; pop_size<=PSO_META_POPSIZE_MAX; pop_size+=PSO_SYS_POPSIZE_STEP)
            for(double cself=PSO_META_CSELF_MIN; cself<=PSO_META_CSELF_MAX; cself+=PSO_SYS_CSELF_STEP)
                for(double cglob=PSO_META_CGLOBAL_MIN; cglob<=PSO_META_CGLOBAL_MAX; cglob+=PSO_SYS_CGLOBAL_STEP)
                    for(double w=PSO_META_INERTIA_MIN; w<=PSO_META_INERTIA_MAX; w+=PSO_SYS_INERTIA_STEP)
                        for(int topology_id=0; topology_id<PSO_META_TOPOLOGY_COUNT; topology_id+=1)
                        {
                            bfunc.local = pso_local_best_functions[topology_id];
                            bfunc.global = pso_global_best_functions[topology_id];
                            if(verbose && !to_outstd)
                                cout << "Testing " << pop_size << " " << cself << " " << cglob << " " << w << " " << topology_id << endl; 
                            out << pop_size << " " << cself << " " << cglob << " " << w << " " << topology_id; 
                            // For each function to compute, call the correct algorithm
                            for( map< NsgaProblem, RESULTS_TYPE>::iterator ii=collected_results.begin(); ii!=collected_results.end(); ++ii)
                            {
                                NsgaProblem p = (*ii).first;
                                // Performs #PROBLEM_REPETITIONS of the selected EA and stores results
                                for(int rep=0; rep < PROBLEM_REPETITIONS; rep++)
                                {
                                    results = ParticleSwarmOptimization(
                                        pop_size,         // population_size,
                                        w, cself, cglob,  // confidence coefficients
                                        p.func_num,       // fitness_function,
                                        &bfunc,           // topology,
                                        p.generations,    // generation_count,
                                        p.evaluations,    // evaluations count
                                        p.stopping,       // use evals or gens
                                        p.fitness_th,     // fitness threshold
                                        p.dimension,      // param_count,
                                        &p.bounds,        // boundaries,
                                        NULL,             // best_individual,
                                        p.ss_improvement, // minimum improvement margin
                                        p.ss_gens         // max number of steady state generation
                                        );
                                    collected_results[(*ii).first]["best_fitness"][rep] = results.best_fitness;
                                    collected_results[(*ii).first]["convergence_eval"][rep] = results.convergence_evals;
                                    collected_results[(*ii).first]["convergence_gen"][rep] = results.convergence_gen;
                                }
                            }
                            collect_results(obj, collected_results, GOALS, PROBLEMS, out);
                            out << endl;
                        }
    }
    outstr.close();
}
