#ifndef __TEST_UTILITIES__
#define __TEST_UTILITIES__

#include <depso/common.h>
#include <depso/test_func.h>
#include <depso/de.h>

#include <string>
#include <vector>
#include <map>
#include <sstream>

#include "parametersParser.h"

using namespace std;

// in a map like this, results will be stored (key is the evaluation item)
typedef map< string, vector<double> > RESULTS_TYPE;

// Returns a string summarizing the active flags
string print_stopping_criterion(StoppingCriterion s)
{
    stringstream ss;
    ss.str("");
    if(s & STOP_EVALS)
        ss << "EVALS ";
    if(s & STOP_FIT)
        ss << "FIT ";
    if(s & STOP_GENS)
        ss << "GENS ";
    if(s & STOP_SS)
        ss << "SS ";
    return ss.str();
}

/// \brief This class represents a problem to be tackled by NSGA-II
class NsgaProblem
{
public:
    /// name of the function
    string func_name;
    /// index of the function
    int func_num;
    /// buondaries of the search space
    DEPSOBoundaries bounds;
    /// problem dimension
    int dimension;
    /// number of generations (if used)
    int generations;
    /// evaluations budget (if used)
    int evaluations;
    /// not used
    int ss_gens;
    /// not used
    double ss_improvement;
    /// fitness threshold (if any)
    double fitness_th;
    /// Stopping criterion
    StoppingCriterion stopping;

    NsgaProblem() : 
        func_name("NOT SET"),
        dimension(0),
        generations(-1),
        evaluations(-1),
        ss_gens(-1),
        ss_improvement(-1),
        fitness_th(-1e9),
        stopping(STOP_NEVER)
    {}

    /// \brief Checks if the Problem is valid.
    
    /// At the current moment, only checks if the stopping criterion has been set.
    bool is_valid()
    {
        if(generations < 0 && evaluations < 0)
            return false;
        return true;
    }

    /// Set the stopping criterion according to the goal to be reached and limits set
    /// to evaluations and generations allowed.
    void set_stopping_criterion(string goal)
    {
        if(evaluations > 0)
            stopping = stopping | STOP_EVALS;
        if(generations > 0)
            stopping = stopping | STOP_GENS;
        if(ss_gens > 0 && ss_improvement > 0)
        {
            stopping = stopping | STOP_SS;
        }
        // if the goal does not request to compute final fitness,
        // you do not have to wait until the end, and you can
        // quit as soon as you reach the fitness threshold
        if(goal.find("fitness") == string::npos)
            stopping = stopping | STOP_FIT;
    }
    
    /// Set the stopping criterion according to the goals to be reached and limits set
    /// to evaluations and generations allowed.
    void set_stopping_criterion(vector<string> goals)
    {
        if(evaluations > 0)
            stopping = stopping | STOP_EVALS;
        if(generations > 0)
            stopping = stopping | STOP_GENS;
        if(ss_gens > 0 && ss_improvement > 0)
            stopping = stopping | STOP_SS;
        // if the goal does not request to compute final fitness,
        // you do not have to wait until the end, and you can
        // quit as soon as you reach the fitness threshold
        bool no_one_needs_fitness = true;
        for(int g=0; g< goals.size(); g++)
            if(goals[g].find_first_of("fitness") == string::npos)
            {
                no_one_needs_fitness = false;
                break;
            }
        if(no_one_needs_fitness)
            stopping = stopping | STOP_FIT;
    }

};

ostream& operator<<(ostream& out, const NsgaProblem& me)
{
    out << "PROBLEM_NAME "       << me.func_name << endl;
    out << "PROBLEM_RANGE_MIN "  << me.bounds.min[0] << endl;
    out << "PROBLEM_RANGE_MAX "  << me.bounds.max[0] << endl;
    out << "PROBLEM_FITNESS_TH " << me.fitness_th << endl;
	out << "PROBLEM_DIMENSION "  << me.dimension << endl;
    if(me.generations > 0)
        out << "PROBLEM_MAX_GENERATIONS " << me.generations << endl;
    if(me.evaluations > 0)
        out << "PROBLEM_MAX_EVALUATIONS " << me.evaluations << endl;
    if(me.ss_improvement > 0)
        out << "PROBLEM_MIN_IMPROVEMENT " << me.ss_improvement  << endl;
    if(me.ss_gens > 0)
        out << "PROBLEM_MAX_STEADY_STATE_GENERATIONS " << me.ss_gens << endl;
    out << "PROBLEM_STOP "       << print_stopping_criterion(me.stopping)  << endl;
    return out;
}

bool operator!= (NsgaProblem &first, NsgaProblem &second)
{
    if(first.func_name != second.func_name)
        return true;
    if(first.dimension != second.dimension)
        return true;
    if(first.generations != second.generations)
        return true;
    if(first.evaluations != second.evaluations)
        return true;
    return false;
}
bool operator== (NsgaProblem &first, NsgaProblem &second)
{
    return !(first != second);
}
bool operator< (const NsgaProblem first, const NsgaProblem second)
{
    if(first.func_name < second.func_name)
        return true;
    if(first.func_name > second.func_name)
        return false;
    if(first.dimension < second.dimension)
        return true;
    if(first.dimension > second.dimension)
        return false;
    if(first.generations < second.generations)
        return true;
    if(first.generations > second.generations)
        return false;
    if(first.evaluations < second.evaluations)
        return true;
    if(first.evaluations > second.evaluations)
        return false;
    return false;
}


int argmax2(double a, double b)
{
    return (a>b) ? 0 : 1;
}

int argmax3(double a, double b, double c)
{
    if(a>b)
        return (a>c) ? 0 : 2;
    return (b>c) ? 1 : 2;
}

template<class T>
T mean(vector<T> v)
{
	T m;
    T sum = 0;
    for(int i=0;i<v.size();i++)
        sum += v[i];
    return sum / (float)(v.size());
}

template<class T>
T median(vector<T> v)
{
	T m;
	size_t size = v.size();
	sort(v.begin(), v.end());
	if (size  % 2 == 0)
		return (v[size / 2 - 1] + v[size / 2]) / 2.0;
	return v[size / 2];
}

template<class T>
T min(vector<T> v)
{
	T m = 1e9;
    for(int i=0;i<v.size();i++)
        if(v[i] < m)
            m = v[i];
    return m;
}

template<class T>
T max(vector<T> v)
{
	T M = 0.0;
    for(int i=0;i<v.size();i++)
        if(v[i] > M)
            M = v[i];
    return M;
}

// look into a configuration file for the problem
// if num_problem < 0 (or not specified) look for single problem
// otherwise for the problem # num_problem
NsgaProblem read_problem_from_file(string param_file, int num_problem=-1)
{
    NsgaProblem p;
    parametersParser pp;
    stringstream ss;
    string tmp_fun;
    double min_bound, max_bound;
    // if num_problem is less than 0, reads a single problem, otherwise
    // reads one of multi problems
    string suffix = ""; //single problem
    if(num_problem >=0) //multi problem
    {
        ss.str("");
        ss << "_" << num_problem;
        suffix = ss.str();
    }
    ss.str(""); ss << "PROBLEM_NAME" << suffix;
    pp.set_parameter(param_file.c_str(), ss.str().c_str(), &(tmp_fun), true);
    p.func_num = get_function_id_by_name(tmp_fun);
    if(p.func_num < 0)
    {
        cout << "Function " << tmp_fun << " still not implemented. Sorry...\n";
        exit(1);
    }
    p.func_name = tmp_fun;
    boundaries_by_function_name(tmp_fun, min_bound, max_bound);
    ss.str("");	ss << "PROBLEM_FITNESS_TH" << suffix;
    pp.set_parameter(param_file.c_str(), ss.str().c_str(), &(p.fitness_th), true);
    ss.str(""); ss << "PROBLEM_DIMENSION" << suffix;
    pp.set_parameter(param_file.c_str(), ss.str().c_str(), &(p.dimension), true);
    ss.str(""); ss << "PROBLEM_MAX_GENERATIONS" << suffix;
    pp.set_parameter(param_file.c_str(), ss.str().c_str(), &(p.generations));
    ss.str(""); ss << "PROBLEM_MAX_EVALUATIONS" << suffix;
    pp.set_parameter(param_file.c_str(), ss.str().c_str(), &(p.evaluations));
    DEInitBoundaries(&(p.bounds), p.dimension, min_bound, max_bound);
    if(!p.is_valid())
    {
        cout << "A problem has (ironically) some problems. Please check!\n" << p << endl;
        exit(1);
    }
    return p;
}


#endif
